//******************************************************************************
// Copyright (c) 2014 Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <stdlib.h>

/* INLINE SPIN
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN
*/

#define MAX_OBJECTS 20
#ifdef CSPIN_FLAG
#define SEEK_SET 0
#define FILE int
#endif

// SPIN OBJECT c : "clibos"
// SPIN OBJECT strsub : "strsubs"
#include "strsubs.h"
// SPIN OBJECT

int debugflag = 0;

/* INLINE SPIN
PUB start(argc, argv)
  c.enter
  strsub.start
  main(argc, argv)
  waitcnt(clkfreq/10+cnt)
  c.exit(0)
*/

void usage(void)
{
    printf("usage: splink file1 file2 file3 [-d]\n");
    exit(1);
}

FILE *OpenFile(char *fname, char *mode)
{
    FILE *fileptr = fopen(fname, mode);

    if (!fileptr)
    {
        printf("Could not open %s\n", fname);
        exit(1);
    }

    return fileptr;
}

int topmets;
int topobjs;
int numobjs;
int offset = 0;
unsigned short objoffs[MAX_OBJECTS];

typedef struct SpinHeaderS {
    int clockfreq;
    int mode;
    int checksum;
    int pbase;
    int vbase;
    int dbase;
    int pcurr;
    int dcurr;
} SpinHeaderT;

unsigned short buf[512];
SpinHeaderT header1;
SpinHeaderT header2;

void PrintHeader(SpinHeaderT *header)
{
    printf("clkfreq = %d, mode = %d, checksum = %d\n", header->clockfreq, header->mode, header->checksum);
    printf("pbase = %d, vbase = %d, dbase = %d, pcurr = %d, dcurr = %d\n",
        header->pbase, header->vbase, header->dbase, header->pcurr, header->dcurr);
}

void GetSpinHeader(FILE *infile, SpinHeaderT *header)
{
    unsigned int buf[4];

    fread(buf, 1, 16, infile);
    header->clockfreq = buf[0];
    header->mode = buf[1] & 255;
    header->checksum = (buf[1] >> 8) & 255;
    header->pbase = buf[1] >> 16;
    header->vbase = buf[2] & 0xffff;
    header->dbase = buf[2] >> 16;
    header->pcurr = buf[3] & 0xffff;
    header->dcurr = buf[3] >> 16;
    if (debugflag) PrintHeader(header);
}

int ComputeCheckSum(char *ptr, int num)
{
    int i;
    int checksum = 0;

    for (i = 0; i < num; i++) checksum += *ptr++;

    return checksum & 255;
}

void WriteHeader(FILE *outfile, SpinHeaderT *header)
{
    unsigned short buf[8];

    buf[0] = header->clockfreq;
    buf[1] = header->clockfreq >> 16;
    buf[2] = (header->checksum << 8) | header->mode;
    buf[3] = header->pbase;
    buf[4] = header->vbase;
    buf[5] = header->dbase;
    buf[6] = header->pcurr;
    buf[7] = header->dcurr;
    fwrite(buf, 1, 16, outfile);
}

void CopyFile(FILE *infile, FILE *outfile, int linkflag)
{
    int i, j;
    int num;

    while ((num = fread(buf, 1, 512, infile)) > 0)
    {
        if (linkflag)
        {
            linkflag = 0;
            i = buf[1] & 255;
            if (debugflag) printf("CopyFile: %d methods\n", i);
            for (j = 0; j < topobjs; j++)
            {
                if (debugflag)
                    printf("CopyFile: Change %d to %d\n", buf[i << 1], objoffs[j]);
                buf[i << 1] = objoffs[j];
                i++;
            }
            if (debugflag)
            {
                for (i = 0; i < 14; i++)
                    printf("%4.4x ", buf[i]);
                printf("\n");
            }
        }
        fwrite(buf, 1, num, outfile);
    }
}

void GetObjectInfo(FILE *infile, int top)
{
    int objnum = 0;
    int num, nummet, numobj, objsize;

    while ((num = fread(buf, 1, 4, infile)) > 0)
    {
        nummet = buf[1] & 255;
        numobj = buf[1] >> 8;
        objsize = buf[0];
        if (debugflag)
        {
            printf("GetObjInfo: obj %3d, methods = %3d, objects = %3d, size = %5d, offset = %5d\n",
                objnum, nummet, numobj, objsize, offset);
        }
        objnum++;
        if (top)
        {
            if (objnum > 1)
            {
                printf("Top object contains more than one object\n");
                exit(1);
            }
            topmets = nummet;
            topobjs = numobj;
        }
        else
        {
            if (objnum > MAX_OBJECTS)
            {
                printf("Too many objects\n");
                exit(1);
            }
            numobjs = objnum;
            objoffs[objnum-1] = offset;
        }
        offset += objsize;
        for (objsize -= 4; objsize > 0; objsize -= num)
        {
            num = 512;
            if (objsize < 512) num = objsize;
            num = fread(buf, 1, num, infile);
            if (num <= 0)
            {
                printf("Object too small\n");
                exit(1);
            }
        }
    }

    fseek(infile, 16, SEEK_SET);
}

int main(int argc, char **argv)
{
    int i;
    FILE *infile1;
    FILE *infile2;
    FILE *outfile;
    int psize1, psize2, vsize2;
    int checksum_old, checksum_new;

    if (argc == 5 && !strcmp(argv[4], "-d"))
        debugflag = 1;
    else if (argc != 4)
        usage();

    infile1 = OpenFile(argv[1], "r");
    infile2 = OpenFile(argv[2], "r");
    outfile = OpenFile(argv[3], "w");

    GetSpinHeader(infile1, &header1);
    GetSpinHeader(infile2, &header2);

    checksum_old = ComputeCheckSum((char *)&header1, sizeof(header1)) - 0x14;
    checksum_old += ComputeCheckSum((char *)&header2, sizeof(header2)) - 0x14;
    checksum_old &= 255;
    if (debugflag) printf("Header checksum = %d\n", checksum_old);
    psize1 = header1.vbase - header1.pbase;
    psize2 = header2.vbase - header2.pbase;
    vsize2 = header2.dbase - header2.vbase - 8;
    if (debugflag)
    {
        printf("psize2 = %d, vsize2 = %d\n", psize2, vsize2);
        printf("psize1 %d\n", psize1);
    }
    header1.vbase += psize2;
    header1.dbase += psize2 + vsize2;
    header1.dcurr += psize2 + vsize2;

    GetObjectInfo(infile1, 1);
    GetObjectInfo(infile2, 0);

    checksum_new = ComputeCheckSum((char *)&header1, sizeof(header1));
    if (debugflag) printf("New checksum = %d\n", checksum_new);

    // Update the checksum
    for (i = 0; i < topobjs; i++)
    {
        checksum_new += ComputeCheckSum((char *)&objoffs[i], 2);
    }

    header1.checksum = (header1.checksum - (checksum_new - checksum_old) + 0x14) & 255;
    if (debugflag) PrintHeader(&header1);

    WriteHeader(outfile, &header1);
    CopyFile(infile1, outfile, 1);
    CopyFile(infile2, outfile, 0);

    fclose(infile1);
    fclose(infile2);
    fclose(outfile);

    return 0;
}
/*
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
*/
